# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import contextlib
import logging
import os
import pathlib
import subprocess
import typing as t

from capellambse import helpers
from capellambse.loader.modelinfo import ModelInfo

from . import abc

LOGGER = logging.getLogger(__name__)


class LocalFileHandler(abc.FileHandler):
    def __init__(
        self,
        path: str | os.PathLike,
        *,
        subdir: str | pathlib.PurePosixPath = "/",
    ) -> None:
        path = pathlib.Path(path, helpers.normalize_pure_path(subdir))

        super().__init__(path)
        self.__transaction: set[pathlib.PurePosixPath] | None = None
        assert isinstance(self.path, pathlib.Path)

    def open(
        self,
        filename: str | pathlib.PurePosixPath,
        mode: t.Literal["r", "rb", "w", "wb"] = "rb",
    ) -> t.BinaryIO:
        assert isinstance(self.path, pathlib.Path)
        normpath = helpers.normalize_pure_path(filename)
        if "w" not in mode or self.__transaction is None:
            path = self.path / normpath
            return path.open("rb")

        if normpath in self.__transaction:
            raise RuntimeError(
                f"File already written in this transaction: {normpath}"
            )
        self.__transaction.add(normpath)
        tmppath = _tmpname(normpath)
        return (self.path / tmppath).open("wb")

    def get_model_info(self) -> ModelInfo:
        assert isinstance(self.path, pathlib.Path)
        if (self.path / ".git").exists():
            return ModelInfo(
                branch=self.__git_rev_parse("--abbrev-ref", "HEAD"),
                title=self.path.name,
                url=self.__git_get_remote_url(),
                rev_hash=self.__git_rev_parse("HEAD"),
            )
        return ModelInfo(title=self.path.name)

    @contextlib.contextmanager
    def write_transaction(
        self, *, dry_run: bool = False, **kw: t.Any
    ) -> t.Generator[t.Mapping[str, t.Any], None, None]:
        """Start a write transaction.

        During the transaction, file writes are redirected to temporary
        files next to the target files, and if the transaction ends
        successfully they are moved to their destinations all at once.

        Parameters
        ----------
        dry_run
            Discard the temporary files after a successful transaction
            instead of committing them to their destinations.
        """
        assert isinstance(self.path, pathlib.Path)

        with super().write_transaction(**kw) as unused_kw:
            if self.__transaction is not None:
                raise RuntimeError("Another transaction is already open")
            self.__transaction = set()

            try:
                yield unused_kw
            except:
                LOGGER.debug("Aborting transaction due to exception")
                dry_run = True
                raise
            finally:
                for file in self.__transaction:
                    tmpname = _tmpname(file)
                    if dry_run:
                        LOGGER.debug("Removing temporary file %s", tmpname)
                        (self.path / tmpname).unlink()
                    else:
                        LOGGER.debug("Committing file %s to %s", tmpname, file)
                        (self.path / tmpname).replace(self.path / file)
                self.__transaction = None

    @property
    def rootdir(self) -> LocalFilePath:
        return LocalFilePath(self, pathlib.PurePosixPath("/"))

    def iterdir(
        self, subdir: str | pathlib.PurePosixPath = "."
    ) -> t.Iterator[LocalFilePath]:
        assert isinstance(self.path, pathlib.Path)
        subdir = helpers.normalize_pure_path(subdir)
        for p in self.path.joinpath(subdir).iterdir():
            yield LocalFilePath(
                self,
                pathlib.PurePosixPath(p.relative_to(self.path)),
            )

    def __git_rev_parse(self, *options: str) -> str | None:
        assert isinstance(self.path, pathlib.Path)
        try:
            return (
                subprocess.run(
                    ["git", "rev-parse", *options],
                    cwd=self.path,
                    check=True,
                    capture_output=True,
                )
                .stdout.decode("utf-8")
                .strip()
            )
        except Exception as err:
            LOGGER.debug(
                "Git rev-parse with options %s failed: %s: %s",
                options,
                type(err).__name__,
                err,
            )
            return None

    def __git_get_remote_url(self) -> str | None:
        assert isinstance(self.path, pathlib.Path)
        try:
            remotes = (
                subprocess.run(
                    ["git", "remote"],
                    cwd=self.path,
                    check=True,
                    capture_output=True,
                )
                .stdout.decode("utf-8")
                .splitlines()
            )
            return (
                subprocess.run(
                    ["git", "remote", "get-url", remotes[0]],
                    cwd=self.path,
                    check=True,
                    capture_output=True,
                )
                .stdout.decode("utf-8")
                .strip()
            )
        except (IndexError, subprocess.CalledProcessError):
            return None


def _tmpname(filename: pathlib.PurePosixPath) -> pathlib.PurePosixPath:
    prefix = "."
    suffix = ".tmp"
    name = filename.name[0 : 255 - (len(prefix) + len(suffix))]
    return filename.with_name(f"{prefix}{name}{suffix}")


class LocalFilePath(abc.AbstractFilePath[LocalFileHandler]):
    def is_dir(self) -> bool:
        base = t.cast(pathlib.Path, self._parent.path)
        path = base.joinpath(self._path).resolve()
        return path.is_dir()

    def is_file(self) -> bool:
        base = t.cast(pathlib.Path, self._parent.path)
        path = base.joinpath(self._path).resolve()
        return path.is_file()



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0
"""Classes that allow access to diagrams in the model."""
from __future__ import annotations

import abc
import base64
import collections.abc as cabc
import importlib.metadata as imm
import logging
import operator
import os
import traceback
import typing as t
import uuid

import markupsafe

import capellambse
from capellambse import aird, diagram, helpers, svg

from . import common as c
from . import modeltypes


@t.runtime_checkable
class DiagramFormat(t.Protocol):
    filename_extension: str

    @classmethod
    def convert(cls, dg: diagram.Diagram) -> t.Any:
        ...

    @classmethod
    def from_cache(cls, cache: bytes) -> t.Any:
        ...


DiagramConverter = t.Union[
    t.Callable[[diagram.Diagram], t.Any],
    DiagramFormat,
]

LOGGER = logging.getLogger(__name__)
REPR_DRAW: bool | None = None


class UnknownOutputFormat(ValueError):
    """An unknown output format was requested for the diagram."""


class AbstractDiagram(metaclass=abc.ABCMeta):
    """Abstract superclass of model diagrams."""

    uuid: str
    """Unique ID of this diagram."""
    name: str
    """Human-readable name for this diagram."""
    target: c.GenericElement
    """This diagram's "target".

    The target of a diagram is usually:

    *   The model element which is the direct parent of all visible
        nodes **OR**
    *   The only top-level element in the diagram **OR**
    *   The element which is considered to be the "element of interest".
    """
    filters: cabc.MutableSet[str]
    """The filters that are activated for this diagram."""

    if t.TYPE_CHECKING:

        @property
        def _allow_render(self) -> bool:
            ...

    else:
        _allow_render: bool = True
        """Allow this diagram to be rendered by the internal rendering engine.

        If this property is set to False, and a diagram cache was
        specified for the model, this diagram can only be loaded from
        the cache, and will never be rendered. Has no effect if there
        was no diagram cache specified.

        :meta public:
        """
    _model: capellambse.MelodyModel
    _render: diagram.Diagram
    _error: BaseException
    _last_render_params: dict[str, t.Any] = {}
    """Additional rendering parameters for the cached rendered diagram.

    Rendering options for :class:`aird.Diagram`s. Handing over
    parameters that differ to these will force a fresh rendering of the
    diagram, flushing the cached diagram.

    The following parameters are currently supported:

        - ``sorted_exchangedItems`` (*bool*): Enable ExchangeItem
          sorting when rendering diagrams with active ExchangeItems
          filter (``show.exchange.items.filter``).
    """

    def __init__(self, model: capellambse.MelodyModel) -> None:
        self._model = model
        self._last_render_params = {}

    def __dir__(self) -> list[str]:
        return dir(type(self)) + [
            f"as_{i.name}"
            for i in imm.entry_points()["capellambse.diagram.formats"]
        ]

    def __getattr__(self, attr: str) -> t.Any:
        if attr.startswith("as_"):
            fmt = attr[len("as_") :]
            try:
                return self.render(fmt)
            except UnknownOutputFormat:
                raise
            except Exception as err:  # pylint: disable=broad-except
                if hasattr(self, "_error") and err is self._error:
                    err_img = self._render
                else:
                    err_img = self.__create_error_image("render", err)
                assert err_img is not None
                converter = _find_format_converter(fmt)
                if isinstance(converter, DiagramFormat):
                    return converter.convert(err_img)
                else:
                    return converter(err_img)
        return getattr(super(), attr)

    def __repr__(self) -> str:
        global REPR_DRAW
        if REPR_DRAW is None:
            REPR_DRAW = TerminalGraphicsFormat.is_supported()

        if not REPR_DRAW:
            return self._short_repr_()

        try:
            escapes = self.render("termgraphics")
        except Exception:
            LOGGER.exception("Cannot render diagram as PNG for display")
            return self._short_repr_()

        return self._short_repr_() + "\n" + escapes.decode("ascii")

    def _short_repr_(self) -> str:
        return f"<Diagram {self.name!r}>"

    def __html__(self) -> markupsafe.Markup:
        return (
            markupsafe.Markup("<figure>")
            + markupsafe.Markup(self.render("svg"))
            + markupsafe.Markup("<figcaption>")
            + self.name
            + markupsafe.Markup("</figcaption></figure>")
        )

    def _short_html_(self) -> markupsafe.Markup:
        return markupsafe.Markup(
            f"<b>{markupsafe.Markup.escape(self.name)}</b> "
            f"(uuid: {markupsafe.Markup.escape(self.uuid)})"
        )

    def _repr_mimebundle_(
        self,
        include: cabc.Container[str] | None = None,
        exclude: cabc.Container[str] | None = None,
        **_,
    ) -> tuple[dict[str, t.Any], dict[t.Any, t.Any]] | dict[str, t.Any] | None:
        if include is None:
            include = helpers.EverythingContainer()
        if exclude is None:
            exclude = ()

        formats: dict[str, DiagramConverter] = {}
        for conv in _iter_format_converters():
            mime = getattr(conv, "mimetype", None)
            if not mime or mime not in include or mime in exclude:
                continue

            formats[mime] = conv
        if not formats:
            return None

        bundle: dict[str, t.Any] = {}
        for mime, conv in formats.items():
            try:
                bundle[mime] = self.__load_cache(conv)
            except KeyError:
                pass

        if bundle:
            return bundle

        render = self.__render_fresh({})
        for mime, conv in formats.items():
            try:
                if isinstance(conv, DiagramFormat):
                    bundle[mime] = conv.convert(render)
                else:
                    bundle[mime] = conv(render)
            except Exception:
                LOGGER.exception("Failed converting diagram with %r", conv)
        if not bundle:
            LOGGER.error("Failed converting diagram for MIME bundle")
            bundle["text/plain"] = repr(self)
        return bundle

    @property
    def nodes(self) -> c.MixedElementList:
        """Return a list of all nodes visible in this diagram."""
        allids = {e.uuid for e in self.render(None) if not e.hidden}
        elems = []
        for elemid in allids:
            assert elemid is not None
            try:
                elem = self._model.by_uuid(elemid)
            except KeyError:
                continue

            elems.append(elem._element)
        return c.MixedElementList(self._model, elems, c.GenericElement)

    @t.overload
    def render(self, fmt: None, /, **params) -> diagram.Diagram:
        ...

    @t.overload
    def render(self, fmt: str, /, **params) -> t.Any:
        ...

    def render(self, fmt: str | None, /, **params) -> t.Any:
        """Render the diagram in the given format."""
        if fmt is not None:
            conv = _find_format_converter(fmt)
        else:

            def conv(i: diagram.Diagram) -> diagram.Diagram:
                return i

        cache_handler = getattr(self._model, "_diagram_cache", None)
        if fmt is not None and cache_handler is not None:
            try:
                return self.__load_cache(conv)
            except KeyError:
                pass

            if not self._allow_render:
                raise RuntimeError(f"Diagram not in cache: {self.name}")

        render = self.__render_fresh(params)
        if isinstance(conv, DiagramFormat):
            return conv.convert(render)
        else:
            return conv(render)

    @abc.abstractmethod
    def _create_diagram(self, params: dict[str, t.Any]) -> diagram.Diagram:
        """Perform the actual rendering of the diagram.

        This method is called by :meth:`render` to perform the actual
        rendering of the diagram. It is passed the parameters that were
        passed to :meth:`render` as a dictionary.

        Subclasses override this method to implement their rendering
        logic. Do not call this method directly, use :meth:`render`
        instead - it will take care of caching and properly converting
        the render output.

        :meta public:
        """

    def __create_error_image(
        self, stage: str, error: Exception
    ) -> diagram.Diagram:
        err_name = (
            "An error occured while rendering diagram\n"
            f"{self.name!r}\n"
            f"(in stage {stage!r})"
        )
        err_msg = (
            "Please report this error to your tools and methods team,"
            " and attach the following information:"
        )
        err_trace = "".join(
            traceback.format_exception(None, error, error.__traceback__)
        )

        diag = diagram.Diagram("An error occured! :(")
        err_box = diagram.Box(
            (200, 0),
            (350, 0),
            label=err_name,
            uuid="error",
            styleclass="Note",
            styleoverrides={
                "fill": diagram.RGB(255, 0, 0),
                "text_fill": diagram.RGB(255, 255, 255),
            },
        )
        diag.add_element(err_box, extend_viewport=False)
        info_box = diagram.Box(
            (200, err_box.pos.y + err_box.size.y + 10),
            (350, 0),
            label=err_msg,
            uuid="info",
            styleclass="Note",
        )
        diag.add_element(info_box, extend_viewport=False)
        trace_box = diagram.Box(
            (0, info_box.pos.y + info_box.size.y + 10),
            (750, 0),
            label=err_trace,
            uuid="trace",
            styleclass="Note",
        )
        diag.add_element(trace_box, extend_viewport=False)
        diag.calculate_viewport()
        return diag

    def __load_cache(self, converter: DiagramConverter):
        cache_handler = getattr(self._model, "_diagram_cache", None)
        cachedir = getattr(self._model, "_diagram_cache_subdir", None)
        if cache_handler is None or cachedir is None:
            raise KeyError(self.uuid)

        if not isinstance(converter, DiagramFormat):
            raise KeyError(self.uuid)

        try:
            ext = converter.filename_extension
            with cache_handler.open(cachedir / (self.uuid + ext)) as f:
                cache = f.read()
        except FileNotFoundError:
            LOGGER.debug("Diagram not in cache: %s (%s)", self.uuid, self.name)
            raise KeyError(self.uuid) from None

        return converter.from_cache(cache)

    def __render_fresh(self, params: dict[str, t.Any]) -> diagram.Diagram:
        # pylint: disable=broad-except
        if not hasattr(self, "_render") or self._last_render_params != params:
            self.invalidate_cache()
            try:
                self._render = self._create_diagram(params)
            except Exception as err:
                self._error = err
                self._render = self.__create_error_image("parse", self._error)

        if hasattr(self, "_error"):
            raise self._error
        return self._render

    def invalidate_cache(self) -> None:
        """Reset internal diagram cache."""
        try:
            del self._render
        except AttributeError:
            pass

        try:
            del self._error
        except AttributeError:
            pass


class Diagram(AbstractDiagram):
    """Provides access to a single diagram."""

    uuid: str = property(
        operator.attrgetter("_element.uid")
    )  # type: ignore[assignment]
    xtype = "viewpoint:DRepresentationDescriptor"
    name: str = property(
        operator.attrgetter("_element.name")
    )  # type: ignore[assignment]
    viewpoint: str = property(
        operator.attrgetter("_element.viewpoint")
    )  # type: ignore[assignment]
    target_uuid: str = property(
        lambda self: self.target.uuid
    )  # type: ignore[assignment]
    """Obsolete."""

    _element: aird.DiagramDescriptor

    @classmethod
    def from_model(
        cls,
        model: capellambse.MelodyModel,
        descriptor: aird.DiagramDescriptor,
    ) -> Diagram:
        """Wrap a diagram already defined in the Capella AIRD."""
        self = cls.__new__(cls)
        self._model = model
        self._element = descriptor
        return self

    def __init__(self, **kw: t.Any) -> None:
        # pylint: disable=super-init-not-called
        raise TypeError("Cannot create a Diagram this way")

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return self._model is other._model and self._element == other._element

    @property
    def nodes(self) -> c.MixedElementList:
        diagram_elements = list(
            aird.iter_visible(self._model._loader, self._element)
        )
        return c.MixedElementList(
            self._model, diagram_elements, c.GenericElement
        )

    @property
    def _allow_render(self) -> bool:
        return self._model._fallback_render_aird

    @property
    def description(self) -> str | None:
        """Return the diagram description."""
        desc = self._element.descriptor.get("documentation")
        return desc and helpers.repair_html(desc)

    @property
    def target(self) -> c.GenericElement:  # type: ignore[override]
        return c.GenericElement.from_model(self._model, self._element.target)

    @property
    def type(self) -> modeltypes.DiagramType:
        """Return the type of this diagram."""
        sc = self._element.styleclass
        try:
            return modeltypes.DiagramType(sc)
        except ValueError:  # pragma: no cover
            LOGGER.warning("Unknown diagram type %r", sc)
            return modeltypes.DiagramType.UNKNOWN

    @property
    def filters(self) -> cabc.MutableSet[str]:
        """Return a set of currently activated filters on this diagram."""
        return aird.ActiveFilters(self._model, self)

    @filters.setter
    def filters(self, filters: cabc.Iterable[str]) -> None:
        self.filters.clear()
        for filter in filters:
            self.filters.add(filter)

    def _create_diagram(self, params: dict[str, t.Any]) -> diagram.Diagram:
        return aird.parse_diagram(self._model._loader, self._element, **params)


class DiagramAccessor(c.Accessor):
    """Provides access to a list of diagrams below the specified viewpoint."""

    def __init__(
        self,
        viewpoint: str | None = None,
        *,
        cacheattr: str | None = None,
    ) -> None:
        super().__init__()
        self.cacheattr = cacheattr
        self.viewpoint = viewpoint

    def __get__(self, obj, objtype=None):
        del objtype
        if obj is None:  # pragma: no cover
            return self

        if self.viewpoint is None:
            descriptors = list(aird.enumerate_diagrams(obj._model._loader))
        else:
            descriptors = [
                d
                for d in aird.enumerate_diagrams(obj._model._loader)
                if d.viewpoint == self.viewpoint
            ]
        return c.CachedElementList(
            obj._model,
            descriptors,
            Diagram,
            cacheattr=self.cacheattr,
        )


class SVGFormat:
    """Convert the diagram to SVG."""

    filename_extension = ".svg"
    mimetype = "image/svg+xml"

    @staticmethod
    def convert(dg: diagram.Diagram) -> str:
        return convert_svgdiagram(dg).to_string()

    @staticmethod
    def from_cache(cache: bytes) -> str:
        return cache.decode("utf-8")


class PNGFormat:
    """Convert the diagram to PNG."""

    filename_extension = ".png"
    mimetype = "image/png"

    @staticmethod
    def convert(dg: diagram.Diagram) -> bytes:
        try:
            import cairosvg
        except OSError as error:
            raise RuntimeError(
                "Cannot import cairosvg. You are likely missing .dll's."
                " Please see the README for instructions."
            ) from error

        return cairosvg.svg2png(SVGFormat.convert(dg))

    @staticmethod
    def from_cache(cache: bytes) -> bytes:
        return cache


def convert_svgdiagram(
    dg: diagram.Diagram,
) -> svg.generate.SVGDiagram:
    """Convert the diagram to a SVGDiagram."""
    jsondata = diagram.DiagramJSONEncoder().encode(dg)
    return svg.generate.SVGDiagram.from_json(jsondata)


class ConfluenceSVGFormat:
    """Convert the diagram to Confluence-style SVG."""

    filename_extension = ".svg"
    prefix = (
        f'<ac:structured-macro ac:macro-id="{uuid.uuid4()!s}" ac:name="html"'
        ' ac:schema-version="1">'
        "<ac:plain-text-body><![CDATA["
    )
    postfix = "]]></ac:plain-text-body></ac:structured-macro>"

    @classmethod
    def convert(cls, dg: diagram.Diagram) -> str:
        return "".join((cls.prefix, SVGFormat.convert(dg), cls.postfix))

    @classmethod
    def from_cache(cls, cache: bytes) -> str:
        return "".join((cls.prefix, cache.decode("utf-8"), cls.postfix))


class SVGDataURIFormat:
    filename_extension = ".svg"
    preamble = "data:image/svg+xml;base64,"

    @classmethod
    def convert(cls, dg: diagram.Diagram) -> str:
        payload = SVGFormat.convert(dg)
        b64 = base64.standard_b64encode(payload.encode("utf-8"))
        return "".join((cls.preamble, b64.decode("ascii")))

    @classmethod
    def from_cache(cls, cache: bytes) -> str:
        b64 = base64.standard_b64encode(cache)
        return "".join((cls.preamble, b64.decode("ascii")))


class SVGInHTMLIMGFormat:
    filename_extension = ".svg"
    mimetype = "text/html"

    @staticmethod
    def convert(dg: diagram.Diagram) -> markupsafe.Markup:
        payload = SVGDataURIFormat.convert(dg)
        return markupsafe.Markup(f'<img src="{payload}"/>')

    @staticmethod
    def from_cache(cache: bytes) -> str:
        payload = SVGDataURIFormat.from_cache(cache)
        return markupsafe.Markup(f'<img src="{payload}"/>')


class TerminalGraphicsFormat:
    """The kitty terminal graphics protocol diagram format.

    This graphics format generates terminal escape codes that transfer
    PNG data to a TTY using the `kitty graphics protocol`__.

    __ https://sw.kovidgoyal.net/kitty/graphics-protocol/
    """

    filename_extension = ".png"

    @classmethod
    def convert(cls, dg: diagram.Diagram) -> bytes:
        data: bytes = PNGFormat.convert(dg)
        return cls.from_cache(data)

    @staticmethod
    def from_cache(cache: bytes) -> bytes:
        container = b"\x1b_Ga=T,q=2,f=100,m=%d;%b\x1b\\"

        chunks: list[bytes] = []
        png_b64 = base64.standard_b64encode(cache)
        while png_b64:
            chunk, png_b64 = png_b64[:4096], png_b64[4096:]
            m = (0, 1)[bool(png_b64)]
            chunks.append(container % (m, chunk))

        return b"".join(chunks)

    @staticmethod
    def is_supported() -> bool:
        """Return whether the used terminal supports graphics.

        This implementation checks whether stdin, stdout and stderr are
        connected to a terminal, and whether the ``$TERM`` environment
        variable is set to a know-supportive value. Currently the only
        recognized value is ``xterm-kitty``.
        """
        return (
            all(os.isatty(i) for i in range(3))
            and os.getenv("TERM") == "xterm-kitty"
        )


def _find_format_converter(fmt: str) -> DiagramConverter:
    try:
        return next(
            i.load()
            for i in imm.entry_points()["capellambse.diagram.formats"]
            if i.name == fmt
        )
    except StopIteration:
        raise UnknownOutputFormat(
            f"Unknown image output format {fmt}"
        ) from None


def _iter_format_converters() -> t.Iterator[DiagramConverter]:
    for ep in imm.entry_points()["capellambse.diagram.formats"]:
        try:
            conv = ep.load()
        except ImportError:
            pass
        else:
            yield conv



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0
"""Implementation of objects and relations for Functional Analysis.

Functional Analysis objects inheritance tree (taxonomy):

.. diagram:: [CDB] FunctionalAnalysis [Taxonomy]

Functional Analysis object-relations map (ontology):

.. diagram:: [CDB] FunctionalAnalysis [Ontology]
"""

from __future__ import annotations

import collections.abc as cabc
import typing as t

from .. import common as c
from .. import modeltypes
from . import capellacommon, capellacore, information, interaction

if t.TYPE_CHECKING:
    from . import cs

XT_COMP_EX_FNC_EX_ALLOC = (
    "org.polarsys.capella.core.data.fa"
    ":ComponentExchangeFunctionalExchangeAllocation"
)
XT_COMP_EX_ALLOC = (
    "org.polarsys.capella.core.data.fa:ComponentExchangeAllocation"
)
XT_CEX_REAL = "org.polarsys.capella.core.data.fa:ComponentExchangeRealization"
XT_FCALLOC = "org.polarsys.capella.core.data.fa:ComponentFunctionalAllocation"


@c.xtype_handler(None)
class ControlNode(c.GenericElement):
    """A node with a specific control-kind."""

    _xmltag = "ownedSequenceNodes"

    kind = c.EnumAttributeProperty(
        "kind", modeltypes.ControlNodeKind, writable=False
    )


@c.xtype_handler(None)
class FunctionRealization(c.GenericElement):
    """A realization that links to a function."""

    _xmltag = "ownedFunctionRealizations"


class AbstractExchange(c.GenericElement):
    """Common code for Exchanges."""

    source = c.AttrProxyAccessor(c.GenericElement, "source")
    target = c.AttrProxyAccessor(c.GenericElement, "target")

    source_port = c.DeprecatedAccessor[c.GenericElement]("source")
    target_port = c.DeprecatedAccessor[c.GenericElement]("target")

    def __dir__(self) -> list[str]:
        attrs = list(super().__dir__())
        attrs.remove("source_port")
        attrs.remove("target_port")
        return attrs


@c.xtype_handler(None)
class AbstractFunction(c.GenericElement):
    """An AbstractFunction."""

    available_in_states = c.AttrProxyAccessor(
        capellacommon.State, "availableInStates", aslist=c.ElementList
    )


@c.xtype_handler(None)
class FunctionPort(c.GenericElement):
    """A function port."""

    owner = c.ParentAccessor(c.GenericElement)
    exchanges: c.Accessor
    state_machines = c.DirectProxyAccessor(
        capellacommon.StateMachine, aslist=c.ElementList
    )


@c.xtype_handler(None)
class FunctionInputPort(FunctionPort):
    """A function input port."""

    _xmltag = "inputs"

    exchange_items = c.AttrProxyAccessor(
        information.ExchangeItem, "incomingExchangeItems", aslist=c.ElementList
    )


@c.xtype_handler(None)
class FunctionOutputPort(FunctionPort):
    """A function output port."""

    _xmltag = "outputs"

    exchange_items = c.AttrProxyAccessor(
        information.ExchangeItem, "outgoingExchangeItems", aslist=c.ElementList
    )


class Function(AbstractFunction):
    """Common Code for Function's."""

    kind = c.EnumAttributeProperty(
        "kind", modeltypes.FunctionKind, default="FUNCTION"
    )

    is_leaf = property(lambda self: not self.functions)

    inputs = c.DirectProxyAccessor(FunctionInputPort, aslist=c.ElementList)
    outputs = c.DirectProxyAccessor(FunctionOutputPort, aslist=c.ElementList)

    exchanges: c.Accessor["FunctionalExchange"]
    functions: c.Accessor
    packages: c.Accessor
    related_exchanges: c.Accessor["FunctionalExchange"]

    realized_functions = c.LinkAccessor[AbstractFunction](
        "ownedFunctionRealizations",
        FunctionRealization,
        aslist=c.MixedElementList,
        attr="targetElement",
    )
    realizing_functions = c.ReferenceSearchingAccessor[AbstractFunction](
        (), "realized_functions", aslist=c.MixedElementList
    )


@c.xtype_handler(None)
class FunctionalExchange(AbstractExchange):
    """A functional exchange."""

    _xmltag = "ownedFunctionalExchanges"

    exchange_items = c.AttrProxyAccessor(
        information.ExchangeItem, "exchangedItems", aslist=c.ElementList
    )

    realized_functional_exchanges = c.LinkAccessor["FunctionalExchange"](
        "ownedFunctionalExchangeRealizations",
        "org.polarsys.capella.core.data.fa:FunctionalExchangeRealization",
        aslist=c.ElementList,
        attr="targetElement",
        backattr="sourceElement",
    )
    realizing_functional_exchanges: c.Accessor["FunctionalExchange"]

    @property
    def owner(self) -> ComponentExchange | None:
        return self.allocating_component_exchange


class FunctionalChainInvolvement(interaction.AbstractInvolvement):
    """Abstract class for FunctionalChainInvolvementLink/Function."""

    _xmltag = "ownedFunctionalChainInvolvements"


@c.xtype_handler(None)
class FunctionalChainInvolvementLink(FunctionalChainInvolvement):
    """An element linking a FunctionalChain to an Exchange."""

    exchanged_items = c.AttrProxyAccessor(
        information.ExchangeItem, "exchangedItems", aslist=c.ElementList
    )
    exchange_context = c.AttrProxyAccessor(
        capellacore.Constraint, "exchangeContext"
    )


@c.xtype_handler(None)
class FunctionalChainInvolvementFunction(FunctionalChainInvolvement):
    """An element linking a FunctionalChain to a Function."""


@c.xtype_handler(None)
class FunctionalChainReference(FunctionalChainInvolvement):
    """An element linking two related functional chains together."""


@c.xtype_handler(None)
class FunctionalChain(c.GenericElement):
    """A functional chain."""

    _xmltag = "ownedFunctionalChains"

    kind = c.EnumAttributeProperty(
        "kind", modeltypes.FunctionalChainKind, default="SIMPLE"
    )

    involvements = c.DirectProxyAccessor(
        c.GenericElement,
        (
            FunctionalChainInvolvementFunction,
            FunctionalChainInvolvementLink,
            FunctionalChainReference,
        ),
        aslist=c.ElementList,
    )
    involved_functions = c.LinkAccessor[AbstractFunction](
        "ownedFunctionalChainInvolvements",
        FunctionalChainInvolvementFunction,
        aslist=c.MixedElementList,
        attr="involved",
    )
    involved_links = c.LinkAccessor[AbstractExchange](
        "ownedFunctionalChainInvolvements",
        FunctionalChainInvolvementLink,
        aslist=c.MixedElementList,
        attr="involved",
    )
    involved_chains = c.LinkAccessor["FunctionalChain"](
        "ownedFunctionalChainInvolvements",
        "org.polarsys.capella.core.data.fa:FunctionalChainReference",
        attr="involved",
        aslist=c.ElementList,
    )
    involving_chains: c.Accessor["FunctionalChain"]

    realized_chains = c.LinkAccessor["FunctionalChain"](
        "ownedFunctionalChainRealizations",
        "org.polarsys.capella.core.data.fa:FunctionalChainRealization",
        attr="targetElement",
        backattr="sourceElement",
        aslist=c.ElementList,
    )
    realizing_chains: c.Accessor["FunctionalChain"]

    control_nodes = c.DirectProxyAccessor(ControlNode, aslist=c.ElementList)

    @property
    def involved(self) -> c.MixedElementList:
        return self.involved_functions + self.involved_links


@c.xtype_handler(None)
class ComponentPort(c.GenericElement):
    """A component port."""

    _xmltag = "ownedFeatures"

    direction = c.EnumAttributeProperty(
        "orientation", modeltypes.FPortDir, writable=False
    )
    owner = c.ParentAccessor(c.GenericElement)
    exchanges: c.Accessor


@c.xtype_handler(None)
class ComponentExchange(AbstractExchange):
    """A functional component exchange."""

    _xmltag = "ownedComponentExchanges"

    kind = c.EnumAttributeProperty(
        "kind", modeltypes.ComponentExchangeKind, default="UNSET"
    )

    allocated_functional_exchanges = c.LinkAccessor[FunctionalExchange](
        None,  # FIXME fill in tag
        XT_COMP_EX_FNC_EX_ALLOC,
        aslist=c.ElementList,
        attr="targetElement",
    )
    allocated_exchange_items = c.AttrProxyAccessor(
        information.ExchangeItem,
        "convoyedInformations",
        aslist=c.ElementList,
    )
    func_exchanges = c.DeprecatedAccessor[FunctionalExchange](
        "allocated_functional_exchanges"
    )

    @property
    def allocating_physical_path(self) -> cs.PhysicalPath | None:
        import warnings

        warnings.warn(
            (
                "allocating_physical_path is deprecated; use"
                " allocating_physical_paths instead, which supports multiple"
                " allocations"
            ),
            DeprecationWarning,
        )

        alloc = self.allocating_physical_paths
        if len(alloc) > 1:
            raise RuntimeError(
                "Multiple allocations; use allocating_physical_paths"
            )
        if not alloc:
            return None
        return alloc[0]

    @property
    def owner(self) -> cs.PhysicalLink | None:
        return self.allocating_physical_link

    @property
    def exchange_items(
        self,
    ) -> c.ElementList[information.ExchangeItem]:
        items = c.ElementList(self._model, [], information.ExchangeItem)
        items.extend(self.allocated_exchange_items)
        func_exchanges = self.allocated_functional_exchanges
        assert isinstance(func_exchanges, cabc.Iterable)
        for exchange in func_exchanges:
            items += exchange.exchange_items
        return items

    def __dir__(self) -> list[str]:
        attrs = list(super().__dir__())
        attrs.remove("func_exchanges")
        return attrs


for _port, _exchange in [
    (ComponentPort, ComponentExchange),
    (FunctionInputPort, FunctionalExchange),
    (FunctionOutputPort, FunctionalExchange),
]:
    c.set_accessor(
        _port,
        "exchanges",
        c.ReferenceSearchingAccessor(
            _exchange, "source", "target", aslist=c.ElementList
        ),
    )
del _port, _exchange

c.set_accessor(
    ComponentExchange,
    "realized_component_exchanges",
    c.LinkAccessor[ComponentExchange](
        "ownedComponentExchangeRealizations",
        XT_CEX_REAL,
        aslist=c.ElementList,
        attr="targetElement",
        backattr="sourceElement",
    ),
)
c.set_accessor(
    ComponentExchange,
    "realizing_component_exchanges",
    c.ReferenceSearchingAccessor(
        ComponentExchange, "realized_component_exchanges", aslist=c.ElementList
    ),
)
c.set_accessor(
    FunctionalExchange,
    "allocating_component_exchange",
    c.ReferenceSearchingAccessor(
        ComponentExchange, "allocated_functional_exchanges"
    ),
)
c.set_accessor(
    FunctionalExchange,
    "realizing_functional_exchanges",
    c.ReferenceSearchingAccessor(
        FunctionalExchange,
        "realized_functional_exchanges",
        aslist=c.ElementList,
    ),
)
c.set_accessor(
    FunctionalExchange,
    "involving_functional_chains",
    c.ReferenceSearchingAccessor(FunctionalChain, "involved_links"),
)

c.set_accessor(
    FunctionalChain,
    "involving_chains",
    c.ReferenceSearchingAccessor(
        FunctionalChain, "involved_chains", aslist=c.ElementList
    ),
)
c.set_accessor(
    FunctionalChain,
    "realizing_chains",
    c.ReferenceSearchingAccessor(
        FunctionalChain, "realized_chains", aslist=c.ElementList
    ),
)

c.set_accessor(
    Function,
    "exchanges",
    c.DirectProxyAccessor(FunctionalExchange, aslist=c.ElementList),
)
c.set_accessor(
    Function,
    "related_exchanges",
    c.ReferenceSearchingAccessor(
        FunctionalExchange,
        "source.owner",
        "target.owner",
        aslist=c.ElementList,
    ),
)
c.set_accessor(
    information.ExchangeItem,
    "exchanges",
    c.ReferenceSearchingAccessor(
        (ComponentExchange, FunctionalExchange),
        "exchange_items",
        "allocated_exchange_items",
        aslist=c.ElementList,
    ),
)



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from .. import common as c


@c.xtype_handler(None)
class Constraint(c.GenericElement):
    """A constraint."""

    _xmltag = "ownedConstraints"

    constrained_elements = c.AttrProxyAccessor(
        c.GenericElement,
        "constrainedElements",
        aslist=c.MixedElementList,
    )

    specification = c.SpecificationAccessor()


@c.xtype_handler(None)
class Generalization(c.GenericElement):
    """A Generalization."""

    _xmltag = "ownedGeneralizations"

    super: c.Accessor


@c.xtype_handler(None)
class EnumerationPropertyLiteral(c.GenericElement):
    """A Literal for EnumerationPropertyType."""

    _xmltag = "ownedLiterals"


@c.xtype_handler(None)
class EnumerationPropertyType(c.GenericElement):
    """An EnumerationPropertyType."""

    _xmltag = "ownedEnumerationPropertyTypes"

    literals = c.DirectProxyAccessor(
        EnumerationPropertyLiteral, aslist=c.ElementList
    )


class PropertyValue(c.GenericElement):
    """Abstract base class for PropertyValues."""

    _xmltag = "ownedPropertyValues"

    enumerations = c.DirectProxyAccessor(
        EnumerationPropertyType, aslist=c.ElementList
    )

    value: c.AttributeProperty | c.AttrProxyAccessor


@c.xtype_handler(None)
class BooleanPropertyValue(PropertyValue):
    """A boolean property value."""

    value = c.BooleanAttributeProperty("value")


@c.xtype_handler(None)
class FloatPropertyValue(PropertyValue):
    """A floating point property value."""

    value = c.AttributeProperty("value", returntype=float, default=0.0)


@c.xtype_handler(None)
class IntegerPropertyValue(PropertyValue):
    """An integer property value."""

    value = c.AttributeProperty("value", returntype=int, default=0)


@c.xtype_handler(None)
class StringPropertyValue(PropertyValue):
    """A string property value."""

    value = c.AttributeProperty("value", default="")


@c.xtype_handler(None)
class EnumerationPropertyValue(PropertyValue):
    """An enumeration property value."""

    type = c.AttrProxyAccessor(EnumerationPropertyType, "type")
    value = c.AttrProxyAccessor(EnumerationPropertyLiteral, "value")


@c.xtype_handler(None)
class PropertyValueGroup(c.GenericElement):
    """A group for PropertyValues."""

    _xmltag = "ownedPropertyValueGroups"

    values = c.DirectProxyAccessor(
        c.GenericElement,
        (
            StringPropertyValue,
            BooleanPropertyValue,
            IntegerPropertyValue,
            FloatPropertyValue,
            EnumerationPropertyValue,
        ),
        aslist=c.ElementList,
        list_extra_args={"mapkey": "name", "mapvalue": "value"},
    )


@c.xtype_handler(None)
class PropertyValuePkg(c.GenericElement):
    """A Package for PropertyValues."""

    _xmltag = "ownedPropertyValuePkgs"

    enumeration_property_types = c.DirectProxyAccessor(
        EnumerationPropertyType, aslist=c.ElementList
    )
    groups = c.DirectProxyAccessor(
        PropertyValueGroup,
        aslist=c.ElementList,
        list_extra_args={"mapkey": "name", "mapvalue": "values"},
    )
    values = c.DirectProxyAccessor(
        c.GenericElement,
        (
            StringPropertyValue,
            BooleanPropertyValue,
            IntegerPropertyValue,
            FloatPropertyValue,
            EnumerationPropertyValue,
        ),
        aslist=c.ElementList,
        list_extra_args={"mapkey": "name", "mapvalue": "value"},
    )


c.set_self_references(
    (PropertyValuePkg, "packages"),
)
c.set_accessor(
    c.GenericElement,
    "constraints",
    c.DirectProxyAccessor(Constraint, aslist=c.ElementList),
)

c.set_accessor(
    c.GenericElement,
    "property_values",
    c.DirectProxyAccessor(
        c.GenericElement,
        (
            BooleanPropertyValue,
            EnumerationPropertyValue,
            FloatPropertyValue,
            IntegerPropertyValue,
            StringPropertyValue,
        ),
        aslist=c.MixedElementList,
        list_extra_args={"mapkey": "name", "mapvalue": "value"},
    ),
)
c.set_accessor(
    c.GenericElement,
    "property_value_groups",
    c.DirectProxyAccessor(
        PropertyValueGroup,
        aslist=c.ElementList,
        list_extra_args={"mapkey": "name", "mapvalue": "values"},
    ),
)

c.set_accessor(
    c.GenericElement,
    "applied_property_values",
    c.AttrProxyAccessor(None, "appliedPropertyValues", aslist=c.ElementList),
)
c.set_accessor(
    c.GenericElement,
    "applied_property_value_groups",
    c.AttrProxyAccessor(
        None, "appliedPropertyValueGroups", aslist=c.ElementList
    ),
)



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0
"""Various diagramming related tools.

This module is used to create diagrams, which can then be exported in
various formats (such as SVG).
"""
# isort: off
from ._vector2d import *

from .capstyle import *
from ._diagram import *
from ._json_enc import *
from ._styleclass import *

import typing as t

if not t.TYPE_CHECKING:
    from ._vector2d import __all__ as _all1
    from .capstyle import __all__ as _all2
    from ._diagram import __all__ as _all3
    from ._json_enc import __all__ as _all4
    from ._styleclass import __all__ as _all5

    __all__ = [*_all1, *_all2, *_all3, *_all4, *_all5]

    del _all1, _all2, _all3, _all4, _all5
del t



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0
"""Functions for receiving the styleclass from a given model object."""
from __future__ import annotations

__all__ = ["get_styleclass"]

import collections.abc as cabc
import re

from capellambse import model


def get_styleclass(obj: model.ModelObject) -> str:
    """Return the styleclass for an individual model object.

    Parameters
    ----------
    obj
        An object received from querying the High-Level API.

    Returns
    -------
    str
        A string used for styling and decorating given ``obj`` in a
        diagram representation.
    """
    styleclass_factory = _STYLECLASSES.get(type(obj).__name__, _default)
    return styleclass_factory(obj)


def _default(obj: model.ModelObject) -> str:
    if isinstance(obj, model.GenericElement):
        return obj.xtype.split(":")[-1]
    return type(obj).__name__


def _association(obj: model.ModelObject) -> str:
    assert isinstance(obj, model.information.Association)
    default_kind = kind = "ASSOCIATION"
    assert isinstance(obj.members, model.ElementList)
    for member in obj.members:
        if member.kind != default_kind:
            kind = member.kind
    return kind.capitalize()


def _component_port(obj: model.ModelObject) -> str:
    assert isinstance(obj, model.fa.ComponentPort)
    if obj.direction is None:
        return "CP_UNSET"
    return f"CP_{obj.direction}"


def _control_node(obj: model.ModelObject) -> str:
    assert isinstance(obj, model.fa.ControlNode)
    return "".join((obj.kind.name.capitalize(), _default(obj)))


def _functional_chain_involvement(obj: model.ModelObject) -> str:
    assert isinstance(
        obj,
        (
            model.fa.FunctionalChainInvolvementLink,
            model.fa.FunctionalChainInvolvementFunction,
        ),
    )
    styleclass = _default(obj)
    if isinstance(obj.parent, model.oa.OperationalProcess):
        return styleclass.replace("Functional", "Operational")
    return styleclass


def _generic_component(obj: model.ModelObject) -> str:
    assert isinstance(obj, model.cs.Component)
    styleclass = _default(obj)
    return "".join(
        (
            styleclass[: -len("Component")],
            "Human" * obj.is_human,
            ("Component", "Actor")[obj.is_actor],
        )
    )


def _physical_component(obj: model.ModelObject) -> str:
    assert isinstance(obj, model.pa.PhysicalComponent)
    styleclass = _generic_component(obj)
    ptrn = re.compile("^(.*)(Component|Actor)$")
    nature = obj.nature.name
    return ptrn.sub(rf"\1{nature.capitalize()}\2", styleclass)


def _part(obj: model.ModelObject) -> str:
    assert isinstance(obj, model.cs.Part)
    assert not isinstance(obj.type, model.ElementList)
    xclass = _default(obj.type)
    if xclass == "PhysicalComponent":
        return _physical_component(obj)
    elif xclass == "Entity":
        return xclass
    return _generic_component(obj.type)


def _port_allocation(obj: model.ModelObject) -> str:
    assert isinstance(obj, model.information.PortAllocation)
    styleclasses = set(
        get_styleclass(p)
        for p in (obj.source, obj.target)
        if not isinstance(p, (model.fa.ComponentPort, model.ElementList))
    )
    return f"{'_'.join(sorted(styleclasses))}Allocation"


def _region(obj: model.ModelObject) -> str:
    assert isinstance(obj, model.capellacommon.Region)
    parent_xclass = _default(obj.parent)
    return f"{parent_xclass}{_default(obj)}"


def _property(obj: model.ModelObject) -> str:
    assert isinstance(obj, model.information.Property)
    return obj.kind.name.capitalize()


_STYLECLASSES: dict[str, cabc.Callable[[model.ModelObject], str]] = {
    "Association": _association,
    "CapellaIncomingRelation": lambda _: "RequirementRelation",
    "CapellaOutgoingRelation": lambda _: "RequirementRelation",
    "ComponentPort": _component_port,
    "ControlNode": _control_node,
    "FunctionalChainInvolvementFunction": _functional_chain_involvement,
    "FunctionalChainInvolvementLink": _functional_chain_involvement,
    "FunctionInputPort": lambda _: "FIP",
    "FunctionOutputPort": lambda _: "FOP",
    "LogicalComponent": _generic_component,
    "Part": _part,
    "PhysicalComponent": _physical_component,
    "PhysicalPort": lambda _: "PP",
    "Property": _property,
    "PortAllocation": _port_allocation,
    "Region": _region,
    "SystemComponent": _generic_component,
}



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0
"""CLI for the diagram cache updating feature."""
from __future__ import annotations

import pathlib
import sys

import capellambse
from capellambse import _diagram_cache, cli_helpers

# pylint: disable=unused-import
from capellambse._diagram_cache import IndexEntry

try:
    import click
except ImportError:

    def _main() -> None:
        """Display a dependency error."""
        print("Missing dependency 'capellambse[cli]'", file=sys.stderr)
        raise SystemExit(1)

else:

    @click.command()
    @click.option(
        "-m",
        "--model",
        "model_",
        type=cli_helpers.ModelCLI(),
        required=True,
        help="Capella model to export diagrams from",
    )
    @click.option(
        "-o",
        "--output",
        type=click.Path(
            file_okay=False, dir_okay=True, path_type=pathlib.Path
        ),
        default="./diagrams",
        help="Directory to store the rendered diagrams in",
        show_default=True,
    )
    @click.option(
        "-f",
        "--format",
        type=click.Choice(sorted(_diagram_cache.VALID_FORMATS)),
        default="svg",
        help="Image output format",
        show_default=True,
    )
    @click.option(
        "--index/--no-index",
        default=True,
        help="Generate index.json and index.html files",
        show_default=True,
    )
    @click.option("--exe", help="Name or path of the Capella executable")
    @click.option("--docker", help="Name of a Docker image containing Capella")
    @click.option(
        "--background/--no-background",
        help="Inserts a white background into the diagrams.",
        default=True,
        show_default=True,
    )
    def _main(
        model_: capellambse.MelodyModel,
        output: pathlib.Path,
        format: str,
        index: bool,
        exe: str | None,
        docker: str | None,
        background: bool,
    ) -> None:
        model_._diagram_cache = capellambse.get_filehandler(output)
        model_._diagram_cache_subdir = pathlib.PurePosixPath(".")

        if docker:
            _diagram_cache.export(
                docker,
                model_,
                format=format,
                index=index,
                force="docker",
                background=background,
            )
        else:
            exe = exe or "capella{VERSION}"
            _diagram_cache.export(
                exe,
                model_,
                format=format,
                index=index,
                force="exe",
                background=background,
            )


if __name__ == "__main__":
    _main()



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0
"""Implementation of objects and relations for Functional Analysis.

Composite Structure objects inheritance tree (taxonomy):

.. diagram:: [CDB] CompositeStructure [Taxonomy]

Composite Structure object-relations map (ontology):

.. diagram:: [CDB] CompositeStructure [Ontology]
"""
from .. import common as c
from . import capellacommon, fa, information

XT_DEPLOY_LINK = (
    "org.polarsys.capella.core.data.pa.deployment:PartDeploymentLink"
)
XT_PHYS_PATH_INV = "org.polarsys.capella.core.data.cs:PhysicalPathInvolvement"
XT_COMPONENT_REALIZATION = (
    "org.polarsys.capella.core.data.cs:ComponentRealization"
)


@c.xtype_handler(None)
class Part(c.GenericElement):
    """A representation of a physical component."""

    _xmltag = "ownedParts"

    type = c.AttrProxyAccessor(c.GenericElement, "abstractType")

    deployed_parts: c.Accessor


@c.xtype_handler(None)
class ExchangeItemAllocation(c.GenericElement):
    """An allocation of an ExchangeItem to an Interface."""

    item = c.AttrProxyAccessor(information.ExchangeItem, "allocatedItem")


@c.xtype_handler(None)
class Interface(c.GenericElement):
    """An interface."""

    exchange_item_allocations = c.DirectProxyAccessor(ExchangeItemAllocation)


@c.xtype_handler(None)
class InterfacePkg(c.GenericElement):
    """A package that can hold interfaces and exchange items."""

    exchange_items = c.DirectProxyAccessor(
        information.ExchangeItem, aslist=c.ElementList
    )
    interfaces = c.DirectProxyAccessor(Interface, aslist=c.ElementList)

    packages: c.Accessor


@c.xtype_handler(None)
class PhysicalPort(c.GenericElement):
    """A physical port."""

    _xmltag = "ownedFeatures"

    owner = c.ParentAccessor(c.GenericElement)
    links: c.Accessor


@c.xtype_handler(None)
class PhysicalLink(PhysicalPort):
    """A physical link."""

    ends = c.PhysicalLinkEndsAccessor(
        PhysicalPort, "linkEnds", aslist=c.ElementList
    )
    linkEnds = c.DeprecatedAccessor[PhysicalPort]("ends")
    exchanges = c.LinkAccessor[fa.ComponentExchange](
        "ownedComponentExchangeAllocations",
        fa.XT_COMP_EX_ALLOC,
        aslist=c.ElementList,
        attr="targetElement",
        backattr="sourceElement",
    )

    physical_paths: c.Accessor

    source = c.IndexAccessor[PhysicalPort]("ends", 0)
    target = c.IndexAccessor[PhysicalPort]("ends", 1)


@c.xtype_handler(None)
class PhysicalPath(c.GenericElement):
    """A physical path."""

    _xmltag = "ownedPhysicalPath"

    involved_items = c.LinkAccessor[c.GenericElement](  # FIXME fill in tag
        None, XT_PHYS_PATH_INV, aslist=c.MixedElementList, attr="involved"
    )
    exchanges = c.LinkAccessor[fa.ComponentExchange](  # FIXME fill in tag
        None, fa.XT_COMP_EX_ALLOC, aslist=c.ElementList, attr="targetElement"
    )

    @property
    def involved_links(self) -> c.ElementList[PhysicalLink]:
        return self.involved_items.by_type("PhysicalLink")


class Component(c.GenericElement):
    """A template class for components."""

    is_abstract = c.BooleanAttributeProperty(
        "abstract", __doc__="Boolean flag for an abstract Component"
    )
    is_human = c.BooleanAttributeProperty(
        "human", __doc__="Boolean flag for a human Component"
    )
    is_actor = c.BooleanAttributeProperty(
        "actor", __doc__="Boolean flag for an actor Component"
    )

    owner = c.ParentAccessor(c.GenericElement)
    state_machines = c.DirectProxyAccessor(
        capellacommon.StateMachine, aslist=c.ElementList
    )
    ports = c.DirectProxyAccessor(fa.ComponentPort, aslist=c.ElementList)
    physical_ports = c.DirectProxyAccessor(PhysicalPort, aslist=c.ElementList)
    parts = c.ReferenceSearchingAccessor(Part, "type", aslist=c.ElementList)
    physical_paths = c.DirectProxyAccessor(PhysicalPath, aslist=c.ElementList)
    physical_links = c.DirectProxyAccessor(PhysicalLink, aslist=c.ElementList)
    exchanges = c.DirectProxyAccessor(
        fa.ComponentExchange, aslist=c.ElementList
    )

    related_exchanges = c.ReferenceSearchingAccessor(
        fa.ComponentExchange,
        "source.owner",
        "target.owner",
        aslist=c.ElementList,
    )

    realized_components = c.LinkAccessor["Component"](
        "ownedComponentRealizations",
        XT_COMPONENT_REALIZATION,
        aslist=c.ElementList,
        attr="targetElement",
    )
    realizing_components = c.ReferenceSearchingAccessor["Component"](
        (), "realized_components", aslist=c.ElementList
    )


@c.xtype_handler(None)
class ComponentRealization(c.GenericElement):
    """A realization that links to a component."""

    _xmltag = "ownedComponentRealizations"


c.set_accessor(
    InterfacePkg,
    "packages",
    c.DirectProxyAccessor(InterfacePkg, aslist=c.ElementList),
)
c.set_accessor(
    Part,
    "deployed_parts",
    c.LinkAccessor(  # FIXME fill in tag
        None, XT_DEPLOY_LINK, aslist=c.ElementList, attr="deployedElement"
    ),
)
c.set_accessor(
    PhysicalPort,
    "links",
    c.ReferenceSearchingAccessor(PhysicalLink, "ends", aslist=c.ElementList),
)
c.set_accessor(
    PhysicalLink,
    "physical_paths",
    c.ReferenceSearchingAccessor(
        PhysicalPath, "involved_items", aslist=c.ElementList
    ),
)
c.set_accessor(
    fa.ComponentExchange,
    "allocating_physical_link",
    c.ReferenceSearchingAccessor(PhysicalLink, "exchanges"),
)
c.set_accessor(
    fa.ComponentExchange,
    "allocating_physical_paths",
    c.ReferenceSearchingAccessor(
        PhysicalPath, "exchanges", aslist=c.ElementList
    ),
)



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0
"""Facilities for working with a native Capella instance."""
from __future__ import annotations

__all__ = ["ARGS_CMDLINE", "native_capella"]

import collections.abc as cabc
import contextlib
import dataclasses
import errno
import json
import logging
import os
import pathlib
import random
import shutil
import string
import subprocess
import tempfile
import typing as t

import capellambse

_LOGGER = logging.getLogger(__name__)

_Runner = t.Union["_SimpleRunner", "_DockerRunner"]
_COMMON_ARGS: t.Final = (
    "--launcher.suppressErrors",
    "-nosplash",
    "-consolelog",
    "-forceimport",
    "-input",
    "main_model",
)
_ARG_WORKSPACE: t.Final = "-data"
_ARG_PROJECT: t.Final = "-import"

ARGS_CMDLINE: t.Final = (
    "-application",
    "org.polarsys.capella.core.commandline.core",
)


@contextlib.contextmanager
def native_capella(
    model: capellambse.MelodyModel,
    *,
    exe: str | pathlib.Path | None = None,
    docker: str | None = None,
) -> cabc.Iterator[_NativeCapella]:
    """Prepare for running a native Capella instance.

    This context manager prepares a temporary workspace and project
    directory, and provides a function for running a native Capella
    instance with the given arguments.

    It currently supports two ways of running a native Capella instance:

    - Using a native executable, by using the ``exe`` argument to
      specify the path to the executable, or the name of the executable
      if it is in the ``PATH``.
    - Using a Docker image, by specifying the image name with the
      ``docker`` argument.

    Both argument values may be strings containing a ``{VERSION}``
    placeholder, which will be replaced with the Capella version of the
    given model. Note that no substitution will be performed if the
    argument is a :class:`pathlib.Path` object.

    Additionally, to facilitate writing wrapper scripts, the process
    specified by the ``exe`` argument is invoked with the following
    additional environment variables:

    - ``CAPELLA_PROJECT``: The path to the project directory
    - ``CAPELLA_WORKSPACE``: The path to the workspace directory
    - ``CAPELLA_VERSION``: The Capella version of the given model

    The context manager yields a callable that can be used to run the
    native Capella instance. The callable takes the arguments to pass
    to Capella as positional arguments, and returns a
    :class:`subprocess.CompletedProcess` object with the result of the
    execution.

    Some cases may require multiple internal subprocess calls, in which
    case either of them may fail. The caller can therefore not rely on a
    :class:`subprocess.CalledProcessError` to belong to the actual
    Capella execution; it may belong to any of the preparation or
    cleanup calls as well. However, the CompletedProcess object returned
    in case of success always belongs to the Capella process.
    """
    with contextlib.ExitStack() as stack:
        workspace = tempfile.TemporaryDirectory(prefix="workspace.")
        workspace_path = pathlib.Path(stack.enter_context(workspace))
        _LOGGER.debug("Creating workspace at %s", workspace_path)

        project = model._loader.write_tmp_project_dir()
        project_path = stack.enter_context(project)
        _LOGGER.debug("Project files written to %s", project_path)

        assert model.info.capella_version
        if exe:
            if isinstance(exe, str):
                exe = exe.replace("{VERSION}", model.info.capella_version)
            runner: _Runner = _SimpleRunner(exe, model.info.capella_version)
        elif docker:
            docker = docker.replace("{VERSION}", model.info.capella_version)
            runner = _DockerRunner(docker)
        else:
            assert False

        yield _NativeCapella(project_path, workspace_path, runner)


@dataclasses.dataclass
class _NativeCapella:
    project: pathlib.Path
    workspace: pathlib.Path
    __runner: _Runner

    def __call__(
        self, *args: str | pathlib.Path
    ) -> subprocess.CompletedProcess:
        try:
            return self.__runner(self.workspace, self.project, *args)
        except subprocess.CalledProcessError as err:
            _LOGGER.error(
                "Native Capella failed with code %d\n%s",
                err.returncode,
                err.stdout,
            )
            raise


class _SimpleRunner:
    def __init__(self, path: str | pathlib.Path, capella_version: str) -> None:
        self.capella_version = capella_version
        path = pathlib.Path(path)
        if path.is_absolute():
            self.path = path
        else:
            if path.parent == pathlib.Path("."):
                exe = shutil.which(path.name)
                if exe is None:
                    raise FileNotFoundError(
                        errno.ENOENT, f"Not found in PATH: {path.name}"
                    )
                self.path = pathlib.Path(exe)
            elif path.exists():
                self.path = path.resolve()
            else:
                raise FileNotFoundError(
                    errno.ENOENT, f"File not found: {path.name}"
                )

    def __call__(
        self,
        workspace: pathlib.Path,
        project: pathlib.Path,
        /,
        *args: str | pathlib.Path,
    ) -> subprocess.CompletedProcess:
        env = os.environ | {
            "CAPELLA_PROJECT": str(project),
            "CAPELLA_WORKSPACE": str(workspace),
            "CAPELLA_VERSION": self.capella_version,
        }
        cmd: list[str | pathlib.Path] = [
            self.path,
            *_COMMON_ARGS,
            _ARG_WORKSPACE,
            workspace,
            _ARG_PROJECT,
            project / "main_model",
            *args,
        ]
        _LOGGER.debug("Running native Capella with command %r", cmd)
        return subprocess.run(
            cmd,
            env=env,
            stdin=subprocess.DEVNULL,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            check=True,
            text=True,
        )


class _DockerRunner:
    def __init__(self, image: str) -> None:
        self.image = image
        self.entrypoint = self.get_entrypoint()

    def __call__(
        self,
        workspace: pathlib.Path,
        project: pathlib.Path,
        /,
        *args: str | pathlib.Path,
    ) -> subprocess.CompletedProcess:
        container_name = "capellambse-" + "".join(
            random.choice(string.ascii_lowercase) for i in range(5)
        )
        with self.start_container(container_name):
            self.copy_files_to_container(
                workspace, "/workspace", container_name
            )
            self.copy_files_to_container(project, "/model", container_name)

            capella = self.run_capella_in_container(
                container_name, self.entrypoint, *args
            )

            self.copy_files_from_container(
                "/workspace/.", workspace, container_name
            )
            self.copy_files_from_container("/model/.", project, container_name)

        return capella

    def get_entrypoint(self) -> list[str]:
        subprocess.run(["docker", "pull", self.image], check=True)
        proc = subprocess.run(
            ["docker", "inspect", self.image],
            check=True,
            stdout=subprocess.PIPE,
            encoding="utf-8",
        )
        parsed_output = json.loads(proc.stdout)[0]
        ep = parsed_output["Config"]["Entrypoint"]
        if not ep:
            raise ValueError(f"Docker image has no entrypoint: {self.image}")
        return ep

    @contextlib.contextmanager
    def start_container(self, name: str) -> cabc.Iterator[None]:
        _LOGGER.debug("Starting container %s from %s", name, self.image)
        with subprocess.Popen(
            [
                "docker",
                "run",
                "--rm",
                "-i",
                "--entrypoint",
                "/bin/sh",
                "--name",
                name,
                self.image,
                "-ec",
                "echo OK; read REPLY || true",
            ],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
        ) as process:
            assert process.stdin is not None is not process.stdout
            try:
                if process.stdout.readline() != b"OK\n":
                    raise RuntimeError(
                        f"Failed handshake with container image {self.image}"
                    )

                _LOGGER.debug("Started container %s", name)
                yield
            finally:
                _LOGGER.debug("Stopping container %s", name)
                process.communicate(b"\n", timeout=30)

    def run_capella_in_container(
        self,
        container_name: str,
        entrypoint: cabc.Sequence[str],
        *args: str | pathlib.Path,
    ) -> subprocess.CompletedProcess:
        return subprocess.run(
            [
                "docker",
                "exec",
                f"{container_name}",
                *entrypoint,
                *_COMMON_ARGS,
                _ARG_WORKSPACE,
                "/workspace",
                _ARG_PROJECT,
                "/model/main_model",
                *args,
            ],
            check=True,
        )

    def copy_files_to_container(
        self, directory: pathlib.Path, destination: str, container_name: str
    ) -> None:
        subprocess.check_call(
            [
                "docker",
                "cp",
                f"{directory}",
                f"{container_name}:{destination}",
            ],
        )

    def copy_files_from_container(
        self, directory: str, destination: pathlib.Path, container_name: str
    ) -> None:
        subprocess.check_call(
            [
                "docker",
                "cp",
                f"{container_name}:{directory}",
                f"{destination}",
            ],
        )



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from ... import common as c
from . import datavalue


@c.xtype_handler(None)
class Enumeration(c.GenericElement):
    """An Enumeration."""

    _xmltag = "ownedDataTypes"

    owned_literals = c.DirectProxyAccessor(
        datavalue.EnumerationLiteral, aslist=c.ElementList
    )

    sub: c.Accessor
    super: c.Accessor[Enumeration]

    @property
    def literals(self) -> c.ElementList[datavalue.EnumerationLiteral]:
        """Return all owned and inherited literals."""
        return (
            self.owned_literals + self.super.literals
            if isinstance(self.super, Enumeration)
            else self.owned_literals
        )



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0
"""Stylesheet generator for SVG diagrams."""
from __future__ import annotations

import collections.abc as cabc
import itertools
import logging
import operator
import re
import typing as t

from capellambse import diagram

logger = logging.getLogger(__name__)
RE_ELMCLASS = re.compile(r"^([A-Z][a-z_]*)(\.[A-Za-z][A-Za-z0-9_]*)?(:.+)?$")
CUSTOM_STYLE_ATTRS = {"marker-fill"}


class Styling:
    """Container for style attributes of svg objects.

    Notes
    -----
    Attributes containing '-' are only referenceable via getattr() or
    subscripting syntax, due to Python identifier naming rules.
    """

    _marker: bool = False

    def __init__(
        self, diagram_class: str | None, class_: str, prefix: str = "", **attr
    ):
        self._diagram_class = diagram_class
        self._class = class_
        self._prefix = prefix

        for key, val in attr.items():
            setattr(self, key, val)

    def __setattr__(self, name: str, value: t.Any) -> None:
        if not name.startswith("_"):
            name = name.replace("_", "-")
        super().__setattr__(name, value)

    def __getattribute__(self, attr: str) -> str:
        if attr in {"marker-start", "marker-end"}:
            defaultstyles = diagram.get_style(self._diagram_class, self._class)
            try:
                value = super().__getattribute__(attr)
            except AttributeError as err:
                try:
                    value = defaultstyles[self._style_name(attr)]
                except KeyError:
                    raise err from None
            try:
                stroke = self.stroke
            except AttributeError:
                stroke = (
                    defaultstyles.get(self._style_name("stroke")) or "#000"
                )
            return f'url("#{value}_{diagram.RGB.fromcss(stroke).tohex()}")'

        return super().__getattribute__(attr)

    def __bool__(self) -> bool:
        try:
            next(iter(self))
        except StopIteration:
            return False
        return True

    def __iter__(self) -> cabc.Iterator[str]:
        defaultstyles = diagram.get_style(self._diagram_class, self._class)
        for attr in ("marker-start", "marker-end"):
            if (
                not self._marker
                and not getattr(super(), attr, None)
                and self._style_name(attr) in defaultstyles
            ):
                yield attr

        yield from itertools.filterfalse(
            operator.methodcaller("startswith", "_"), dir(self)
        )

    def __getitem__(self, attrs: str | cabc.Iterable[str]) -> str | None:
        if isinstance(attrs, str):
            attrs = (attrs,) if attrs else self
        return (
            "; ".join(f"{a}: {self._to_css(getattr(self, a))}" for a in attrs)
            or None
        )

    @classmethod
    def _to_css(
        cls, value: float | int | str | diagram.RGB | cabc.Iterable | None
    ) -> float | int | str:
        if isinstance(value, (str, int, float)):
            return value
        elif value is None:
            return "none"
        elif isinstance(value, diagram.RGB):
            return f"#{value.tohex()}"
        elif isinstance(value, cabc.Iterable):
            return f'url("#{cls._generate_id("CustomGradient", value)}")'
        raise ValueError(f"Invalid styling value: {value!r}")

    @staticmethod
    def _generate_id(
        name: str, value: cabc.Iterable[str | diagram.RGB]
    ) -> str:
        """Return unqiue identifier for given css-value."""
        return "_".join(
            itertools.chain(
                (name,),
                (diagram.RGB.fromcss(v).tohex() for v in value),
            ),
        )

    def __str__(self) -> str:
        return self[""] or ""

    def _style_name(self, attr: str) -> str:
        if not self._prefix:
            return attr

        return "_".join((self._prefix, attr))

    def __contains__(self, obj: t.Any) -> bool:
        return hasattr(self, obj)
